import random
import logging

logger = logging.getLogger(__name__)
name = "Heartbeat"

start_message = name + " started!"

description = "La-dub la-dub"

schema = {
    "color": {
        "value": {"type": "color", "default": [255, 0, 0]},
        "user_input": True,
        "required": False,
    }
}


def update(lights, step, state):
    color = state["color"].get_color(step)
    sub_step = step % 60
    lub = sub_step in range(0, 3) or sub_step in range(9, 12)
    brightness = 1 if lub else 0.4
    lights.set_all_pixels(*tuple(map(lambda x: int(x * brightness), color)))



# This is what will appear in all interfaces
name = "On"

# This is what the user will see after the effect starts
start_message = "The lights have been turned on!"

# This is what will appear in tips and help menus
description = "Turns all of the lights on to a specfied color"

default_speed = 0

# This defines the format of update's 'state' parameter
schema = {
    "color": {
        "value": {"type": "color", "default": (255, 255, 255)},
        "user_input": True,
        "required": False,
    },
}

# This is the function that updates the effect.
# Params:
#   lights: A reference to the light controls (the only way to make anything happen).
#   step: The number of times that this effect has been updated
#   state: Dict with information about the state of the effect
def update(lights, step, state):
    color = state["color"].get_color(step)
    lights.set_all_pixels(*color)



class History:
    def __init__(self, commands):
        self.ordered_events = []
        self.current_index = 0
        self.commands = commands

    def save(self, state):
        self.ordered_events.append(HistoryEvent(state))
        self.current_index = len(self.ordered_events) - 1

    def back(self):
        if self.current_index > 0:
            self.current_index -= 1
            self.ordered_events[self.current_index].restore(self.commands)
            return True
        else:
            return False

    def forward(self):
        if self.current_index < len(self.ordered_events) - 1:
            self.current_index += 1
            self.ordered_events[self.current_index].restore(self.commands)
            return True
        else:
            return False


class HistoryEvent:
    def __init__(self, state):
        self.state = state

    def restore(self, commands):
        commands.stop_all()
        for effect in self.state:
            properties = {
                "ranges": effect["sections"],
                "opacity": effect["opacity"],
                "overlayed": True,
            }
            commands._start_effect(
                effect["effect_name"],
                effect["effect_state"],
                properties,
                effect["transaction_id"],
            )



#! /usr/bin/python3
import argparse
import json
import logging
import socket
import sys

logger = logging.getLogger(__name__)


def start_effect(effect_name, effect_args=None, properties=None):
    s = None
    if effect_args is None:
        effect_args = {}
    try:
        s = socket.socket(socket.AF_UNIX)
        s.connect("/tmp/litd")
        request = {
            "start": {
                "effect": {
                    "name": effect_name,
                    "args": effect_args,
                    "properties": properties,
                }
            }
        }
        s.sendall(json.dumps(request).encode())
    except Exception as e:
        s.close()
        raise conn_error(e)

    res = get_response(s)
    s.close()
    # rc 0: success, rc 2: command usage error
    if res.get("code", 1) != 0:
        logger.error(response_error(res))
    return res


def start_preset(preset, properties=None):
    s = None
    try:
        s = socket.socket(socket.AF_UNIX)
        s.connect("/tmp/litd")
        request = {"start": {"preset": {"name": preset, "properties": properties}}}
        s.sendall(json.dumps(request).encode())
    except Exception as e:
        s.close()
        raise conn_error(e)

    res = get_response(s)
    s.close()
    # rc 0: success, rc 2: command usage error
    if res.get("code", 1) != 0:
        logger.error(response_error(res))
    return res


def stop(effect_id=None, transaction_id=None):
    s = None
    try:
        s = socket.socket(socket.AF_UNIX)
        s.connect("/tmp/litd")
        if effect_id is not None:
            request = {"stop": {"effect_id": effect_id}}
        elif transaction_id is not None:
            request = {"stop": {"transaction_id": transaction_id}}
        else:
            raise ValueError("effect_id or transaction_id must not be None")

        s.sendall(json.dumps(request).encode())
    except Exception as e:
        s.close()
        raise conn_error(e)

    res = get_response(s)
    s.close()
    # rc 0: success, rc 2: command usage error
    if res.get("code", 1) != 0:
        logger.error(response_error(res))
    return res


def query(query):
    s = None
    try:
        s = socket.socket(socket.AF_UNIX)
        s.connect("/tmp/litd")
        msg = {"query": {"what": query}}
        s.sendall(json.dumps(msg).encode())
    except Exception as e:
        s.close()
        raise conn_error(e)

    res = get_response(s)
    s.close()
    if res.get("code", 1) != 0:
        logger.error(response_error(res))
    return res


def back():
    s = None
    try:
        s = socket.socket(socket.AF_UNIX)
        s.connect("/tmp/litd")
        msg = {"start": {"history": {"back": 1}}}
        s.sendall(json.dumps(msg).encode())
    except Exception as e:
        s.close()
        raise conn_error(e)

    res = get_response(s)
    s.close()
    if res.get("code", 1) != 0:
        logger.error(response_error(res))
    return res


def forward():
    s = None
    try:
        s = socket.socket(socket.AF_UNIX)
        s.connect("/tmp/litd")
        msg = {"start": {"history": {"forward": 1}}}
        s.sendall(json.dumps(msg).encode())
    except Exception as e:
        s.close()
        raise conn_error(e)

    res = get_response(s)
    s.close()
    if res.get("code", 1) != 0:
        logger.error(response_error(res))
    return res


def dev_command(command, args):
    s = None
    try:
        s = socket.socket(socket.AF_UNIX)
        s.connect("/tmp/litd")
        msg = {"dev": {"command": command, "args": args}}
        s.sendall(json.dumps(msg).encode())
    except Exception as e:
        s.close()
        raise conn_error(e)

    res = get_response(s)
    s.close()
    if res.get("code", 1) != 0:
        logger.error(response_error(res))
    return res


def get_response(s):
    response = ""
    try:
        # First 32 bytes are a string representation of the message length
        expected = int(s.recv(32).decode())
        received = 0
        while received < expected:
            msg = s.recv(1024).decode()
            received += 1024
            response += msg
    except Exception as e:
        raise conn_error(e)
    return json.loads(response)


def conn_error(e):
    return Exception("LIT Daemon error: {}".format(e))


def response_error(res):
    return Exception(
        "Daemon returned a non-zero response code={}. Error={}".format(
            res.get("code", None), res.get("message", "No message")
        )
    )


def get_effects():
    return query("effects")["effects"]


def get_presets():
    return query("presets")["presets"]


def get_colors():
    return query("colors")["colors"]


def get_color_types():
    return query("color_types")["color_types"]


def get_speeds():
    return query("speeds")["speeds"]


def get_sections():
    return query("sections")["sections"]


def get_zones():
    return query("zones")["zones"]


def get_pixels():
    return query("pixels")["pixels"]


def get_state():
    return query("state")["state"]


if __name__ == "__main__":
    print("This is a library that must be imported to be used")



import random
import logging

logger = logging.getLogger(__name__)
name = "Fireflies"

start_message = name + " started!"

description = "You would not believe your eyes"


def setup_start_duration_colors(lights, args):
    return [(-1, 1, (0, 0, 0))] * lights.size


schema = {
    "color": {
        "value": {"type": "color", "default": (255, 240, 0)},
        "user_input": True,
        "required": False,
    },
    "density": {
        "value": {"type": "number", "min": 1, "max": 100, "default": 50},
        "user_input": True,
        "required": False,
    },
    "start_duration_colors": {
        "value": {
            "type": "(int, int, (int, int, int)) list",
            "default_gen": setup_start_duration_colors,
        },
        "user_input": False,
    },
}

# Average time a led is on
AVG_ON_TIME = 100
MIN_ON_TIME = 10


def update(lights, step, state):
    color_gen = state["color"]
    start_duration_colors = state["start_duration_colors"]
    """ MATH:
    DENSITY = OnTime/(Ontime+OffTime)
    OnTime(1-DENSITY) = Density*OffTime
    Offtime = ((1-Density)*OnTime)/Density 
    """
    density = state["density"] / 100
    avg_off_time = ((1 - density) * AVG_ON_TIME) / density
    for i in range(lights.size):
        sdc = start_duration_colors[i]
        if sdc[0] + sdc[1] <= step:
            state["start_duration_colors"][i] = (
                step + random.random() * int(2 * avg_off_time),
                MIN_ON_TIME + (random.random() * (AVG_ON_TIME - MIN_ON_TIME) * 2),
                color_gen.get_color(step, i),
            )
        brightness = (step - sdc[0]) / (sdc[1] / 2)
        if brightness > 1:
            brightness = 2 - brightness
        brightness = max(brightness, 0)
        lights.set_pixel(
            i,
            int(sdc[2][0] * brightness),
            int(sdc[2][1] * brightness),
            int(sdc[2][2] * brightness),
            brightness,
        )



import logging

logger = logging.getLogger(__name__)

# This is what will appear in all interfaces
name = "Effect Display Name"

# This is what the user will see after the effect starts
start_message = name + " started!"

# This is what will appear in tips and help menus
description = "A light effect"

# Schema defines what information the effect needs. This includes all arguments
# that the effect requires as well as any other state that the effect may need.
# All keys in the 'schema' dict will have a corresponding key in the 'state' dict below.
# All schema elements must have a 'value' and 'user_input' field.
# 'value' defines the what data should look like. The fields are:
#'type': {'number', 'color'} required if user_input=True
#'min': minimum allowed value. required if type=number
#'max': maximum allowed value. required if type=number
# *One of the following two fields are required
#'default': The default value. Must be a primitive
#'default_gen': A function that returns a default value.
# 'user_input': True if this field is meant to be an argument passed by the user

schema = {
    "speed": {
        "value": {"type": "number", "min": 0, "max": 100, "default": 50},
        "user_input": True,
    },
    "color": {"value": {"type": "color", "default": (255, 0, 255)}, "user_input": True},
    "custom": {"value": {"default_gen": lambda x: list()}, "user_input": False},
}

# default_speed is the update frequency (in hertz) that the update function will be called
# if no other rate is specified.
default_speed = 20

# This is the function that controls the effect. Look at the included effects for examples.
# Params:
#    lights: A reference to the light controls (A LED_Controller object)
#    step: The number of iterations since starting the effect
#    state: A dictionary that corresponds to the schema above
def update(lights, step, state):
    # Called repeatedy, with step increasing by 1 each call and state persisting
    state["my_list"].append(step)
    lights.set_all_pixels(*state["color"])



import logging

logger = logging.getLogger(__name__)


class Section:
    def __init__(self, name, start_index, size, section_adapter):
        self.name = name
        self.start_index = start_index
        self.size = size
        self.end_index = start_index + size
        self.section_adapter = section_adapter
        self.absolute_range = range(start_index, self.end_index)


class SectionAdapter:
    """ A SectionAdapter represents a Section's view of its display device.
    Each Section has it's own SectionAdapter
    Indexing as relative to the Section
    """

    def __init__(self, offset, display_adapter):
        self.offset = offset
        self.display_adapter = display_adapter

    def set_pixel_color_rgb(self, n, r, g, b):
        self.display_adapter.set_pixel_color_rgb(n + self.offset, r, g, b)



import colorsys
import random

name = "Random Colors"
description = "Random fully saturated colors"
schema = {}


def get_generator(args):
    def generator(_step, _position):
        return list(
            map(lambda x: int(255 * x), colorsys.hsv_to_rgb(random.random(), 1, 1))
        )

    return generator



import random

name = "Collision"

start_message = name + " started!"

description = "Two projectiles are sent from opposite ends and collide in the center"


def create_projectiles(lights, args):
    return [[0, 1, random.random()], [lights.size - 1, -1, random.random()]]


schema = {
    "projectiles": {
        "value": {"type": "tuple list", "default_gen": create_projectiles},
        "user_input": False,
    },
}

TRAIL_LENGTH = 15


def update(lights, step, state):
    # NOTE WIP
    projectiles = state["projectiles"]
    if random.random() < 0.01:
        projectiles.append([0, 1, random.random()])
        projectiles.append([lights.size - 1, -1, random.random()])
    for i, v in enumerate(projectiles):
        projectiles[i][0] = projectiles[i][0] + projectiles[i][1]
        for t in range(0, TRAIL_LENGTH):
            if v[0] - v[1] * t >= 0 and v[0] - v[1] * t < lights.size:
                lights.set_pixel_hsv(
                    v[0] - v[1] * t, v[2], 1, (TRAIL_LENGTH - 1.0 * t) / TRAIL_LENGTH
                )
    for i1, p1 in enumerate(projectiles):
        for i2, p2 in enumerate(projectiles):
            if p1 is not p2:
                prev1 = p1[0] - p1[1]
                prev2 = p2[0] - p2[1]
                if (prev1 - prev2 < 0) != (p1[0] - p2[0] < 0):
                    # explode(lights, p1[0], p1[2], p2[2]) TODO
                    projectiles.remove(p1)
                    projectiles.remove(p2)


def explode(lights, n, color1, color2):
    radius = lights.size // 20
    for d in range(0, radius):
        for pixel in range(n - d, n + d):
            if pixel >= 0 and pixel < lights.size:
                lights.set_pixel_hsv(
                    pixel,
                    random.choice([color1, color2]),
                    1,
                    (1.0 * radius - d) / radius,
                )



import itertools
import time

DEFAULT_SPEED = 30  # hertz
EFFECT_IDS = itertools.count(0, 1)


class Effect:
    def __init__(self, module):
        self.name = getattr(module, "name", "Unnamed")
        self.schema = getattr(module, "schema", {})
        self.start_message = getattr(module, "start_message", f"{self.name} started")
        self.command_schema = {
            k: {k2: v2 for k2, v2 in v.items() if k2 != "user_input"}
            for k, v in self.schema.items()
            if v.get("user_input", False)
        }
        self.default_speed = getattr(module, "default_speed", DEFAULT_SPEED)
        self.module = module

    def create(self, args, speed, controller, transaction_id):
        return EffectInstance(self, args, speed, controller, transaction_id)

    def __str__(self):
        return str(
            {
                "name": self.name,
                "schema": self.schema,
                "default_speed": self.default_speed,
            }
        )


class EffectInstance:
    def __init__(self, effect, initial_state, speed, controller, transaction_id):
        self.effect = effect
        self.state = initial_state.copy()
        self.step = 0
        self.speed = speed
        self.next_upd_time = time.time()
        self.id = next(EFFECT_IDS)
        self.controller = controller
        self.transaction_id = transaction_id
        self.static_pixels = []

    def update(self):
        # Speed is in units of updates/second
        if self.speed > 0:
            self.effect.module.update(self.controller, self.step, self.state)
            self.step += 1
            self.next_upd_time += 1 / self.speed
        else:
            if not self.static_pixels:
                self.effect.module.update(self.controller, self.step, self.state)
            self.static_pixels = self.controller.get_pixels()
            self.controller.set_pixels(self.static_pixels)
            self.next_upd_time += 1 / DEFAULT_SPEED


    def __str__(self):
        return str(
            {
                "name": self.effect.name,
                "state": self.state,
                "default_speed": self.speed,
            }
        )



import colorsys
import random

name = "Chase"

start_message = name + " started!"

description = (
    "The string is sequentially covered by random colors"
)


def create_heads(lights, args):
    num_heads = int(args["number"])
    colors = [random.random() for _ in range(0, num_heads)]
    head_locations = [i * (-lights.size // num_heads) for i in range(0, num_heads)]
    return [head_locations, colors]


schema = {
    "number": {
        "value": {
            "type": "number",  # TODO integer type (or step)
            "min": 1,
            "max": 20,
            "default": 3,
        },
        "user_input": True,
        "required": False,
        "index": 0,
    },
    "color": {
        "value": {
            "type": "color",  # TODO integer type (or step)
            "default": {"type": "Random Colors"},
        },
        "user_input": True,
        "required": False,
    },
    "heads": {
        "value": {"type": "tuple list", "default_gen": create_heads},
        "user_input": False,
    },
}


def update(lights, step, state):
    heads = state["heads"]
    for i in range(len(heads[0])):
        if heads[0][i] >= 0:
            lights.set_pixel_hsv(heads[0][i], heads[1][i], 1, 1)
        heads[0][i] += 1
        if heads[0][i] >= lights.size:
            heads[0][i] = 0
            heads[1][i] = colorsys.rgb_to_hsv(state["color"].get_color(step))



name = "Throb"

start_message = name + " started!"

description = "Shifts through brighness levels repeatedly"

schema = {
    "color": {
        "value": {"type": "color", "default": (255, 0, 0)},
        "user_input": True,
        "required": False,
    },
    "brightness": {"value": {"type": "number", "default": 0}, "user_input": False},
    "delta": {"value": {"type": "number", "default": 0.05}, "user_input": False},
}


def update(lights, step, state):
    brightness = state["brightness"]
    color = state["color"].get_color(step)
    lights.set_all_pixels(
        int(color[0] * brightness),
        int(color[1] * brightness),
        int(color[2] * brightness),
    )
    if brightness + state["delta"] > 1 or brightness + state["delta"] < 0:
        state["delta"] *= -1
    state["brightness"] += state["delta"]



from .lit import (
    start_effect,
    start_preset,
    stop,
    forward,
    back,
    query,
    dev_command,
    get_effects,
    get_presets,
    get_colors,
    get_color_types,
    get_sections,
    get_zones,
    get_pixels,
    get_state,
)



name = "Static Color"
description = "A constant color"
schema = {"color": {"value": {"type": "rgb"}, "user_input": True, "required": True,}}


def get_generator(args):
    def generator(_step, _position):
        return args["color"]

    return generator



name = "Pride"

start_message = name + " started!"

description = "Displays rainbow pride flag colors"

COLORS = [
    (231, 0, 0),
    (255, 85, 0),
    (255, 239, 0),
    (0, 129, 0),
    (0, 68, 255),
    (74, 0, 118),
]

default_speed = 1

schema = {}


def update(lights, step, state):
    stripe_width = lights.size / len(COLORS)
    start = 0
    end = stripe_width
    for ci in range(len(COLORS)):
        for i in range(int(start), int(end)):
            lights.set_pixel(i, *COLORS[(ci + step) % len(COLORS)])
        start += stripe_width
        end += stripe_width



name = "Disco"

start_message = name + " started!"

description = "Color spectrum is repeatedly 'squished' to one side."

schema = {}


def update(lights, step, state):
    off = step / 10
    for i in range(lights.size):
        lights.set_pixel_hsv(i, ((i * off) / float(lights.size)) % 1, 1, 1)



import logging
import time
import socket

logger = logging.getLogger(__name__)


class DeviceAdapter:
    """ A DeviceAdapter represents a physical display device.
    Many Sections can share one DeviceAdapter 
    Indexing is relative to the device
    """

    def __init__(self, name, size):
        self.size = size
        self.name = name

    @staticmethod
    def from_config(config):
        adapter_type = config.get("type")
        if adapter_type == "ws2812":
            adapter = WS2812Adapter(config["name"], config["size"], config["pin"])
        elif adapter_type == "udp":
            adapter = UDPAdapter(
                config["name"], config["size"], config["ip"], config["port"]
            )
        return adapter

    def set_pixel_color_rgb(self, n, r, g, b):
        pass

    def show(self):
        pass

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        # Device adapters are unique by name
        return hash(self.name)


class UDPAdapter(DeviceAdapter):
    def __init__(self, name, size, ip, port):
        super().__init__(name, size)
        self.ip = ip
        self.port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.pixels = [0] * size * 3

    def set_pixel_color_rgb(self, n, r, g, b):
        self.pixels[3 * n] = r
        self.pixels[3 * n + 1] = g
        self.pixels[3 * n + 2] = b

    def show(self):
        payload = bytearray(self.pixels) + int(time.time() * 100).to_bytes(8, "little")
        self.socket.sendto(payload, 0, (self.ip, self.port))


class WS2812Adapter(DeviceAdapter):

    created = False

    def __init__(self, name, size, pin):
        if WS2812Adapter.created:
            raise Exception("There can only be one adapter of type ws2812")
        super().__init__(name, size)
        self.pin = pin
        self.size = size
        self.ws2812 = self._init_ws2812()
        WS2812Adapter.created = True

    def set_pixel_color_rgb(self, n, r, g, b):
        super().set_pixel_color_rgb(n, r, g, b)
        self.ws2812.setPixelColorRGB(n, r, g, b)

    def show(self):
        self.ws2812.show()

    def _init_ws2812(self):
        from rpi_ws281x import Adafruit_NeoPixel, WS2812_STRIP

        LED_FREQ_HZ = 800000  # LED signal frequency in hertz (usually 800khz)
        LED_DMA = 10  # DMA channel to use for generating signal (try 5)
        LED_BRIGHTNESS = 255  # Set to 0 for darkest and 255 for brightest
        # True to invert the signal (when using NPN transistor level shift)
        LED_INVERT = False
        LED_CHANNEL = 0

        LED_STRIP = WS2812_STRIP
        ws2812 = Adafruit_NeoPixel(
            self.size,
            self.pin,
            LED_FREQ_HZ,
            LED_DMA,
            LED_INVERT,
            LED_BRIGHTNESS,
            LED_CHANNEL,
            LED_STRIP,
        )
        ws2812.begin()
        return ws2812



name = "Fade"

start_message = "Fading!"

description = "Lights fade from one color to another"

TOTAL_STEPS = 1000


def calc_delta(lights, args):
    ar, ag, ab = args["start color"]
    br, bg, bb = args["end color"]
    return ((br - ar) / TOTAL_STEPS, (bg - ag) / TOTAL_STEPS, (bb - ab) / TOTAL_STEPS)


schema = {
    "start color": {
        "value": {"type": "color", "default": (0, 0, 0)},
        "user_input": True,
        "required": False,
        "index": 0,
    },
    "end color": {
        "value": {"type": "color", "default": (255, 255, 255)},
        "user_input": True,
        "required": False,
        "index": 0,
    },
    "delta": {
        "value": {"type": "number tuple", "default_gen": calc_delta},
        "user_input": False,
    },
}


def update(lights, step, state):
    if step >= TOTAL_STEPS:
        # TODO hotswap effects
        end_color = state["end color"].get_color()
        lights.set_all_pixels(*end_color)
        return
    ar, ag, ab = state["start color"].get_color()
    dr, dg, db = state["delta"]
    nr = ar + dr * step
    ng = ag + dg * step
    nb = ab + db * step
    lights.set_all_pixels(max(0, int(nr)), max(0, int(ng)), max(0, int(nb)))



#! /usr/bin/python3
import argparse
import itertools
import json
import logging
import os
import operator
import shutil
import socket
import sys
import threading

from . import commands

MAX_CONNECTIONS = 5
TRANSACTION_IDS = itertools.count(0, 1)
SOCKET_PATH = "/tmp/litd"
logger = logging.getLogger(__name__)


class LitDaemon:
    def __init__(self):
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s:%(name)s:(%(lineno)d):%(levelname)s:%(message)s",
        )
        parser = argparse.ArgumentParser(description="Start the L.I.T. daemon")
        parser.add_argument(
            "--config",
            "-c",
            dest="base_path",
            type=str,
            help="specify the directory containing the config directory",
        )
        parser.add_argument(
            "--gen-config",
            "-g",
            dest="gen_config_path",
            type=str,
            help="specify the new config directory ",
        )
        args = parser.parse_args()

        default_base_path = os.path.dirname(os.path.abspath(__file__))
        if args.gen_config_path:
            try:
                shutil.copytree(
                    os.path.join(default_base_path, "config"),
                    os.path.join(args.gen_config_path, "config"),
                )
                print("Created config directory: {}".format(args.gen_config_path))
                sys.exit(0)
            except Exception as e:
                print("Could not copy config directory: {}".format(e))
                sys.exit(1)

        logger.info("setting up")
        base_path = (
            args.base_path
            if args.base_path and os.path.isdir(args.base_path)
            else default_base_path
        )
        if args.base_path:
            # Create effects symlink to config path
            try:
                src = os.path.join(base_path, "effects")
                dest = os.path.join(os.path.dirname(__file__), "effects", "user")
                if not os.path.exists(dest) and os.path.exists(src):
                    os.symlink(src, dest)
                    logger.info(
                        "Created user effects symlink {} -> {}".format(src, dest)
                    )
                    logger.warning(
                        "User effects will not be available until this program is restarted"
                    )
            except OSError as e:
                logger.warning("Could not create user effects symlink: {}".format(e))
        self.commands = commands.commands(base_path=base_path)
        self.queries = {
            "effects": self.effects,
            "presets": self.presets,
            "colors": self.colors,
            "color_types": self.color_types,
            "sections": self.sections,
            "zones": self.zones,
            "pixels": self.pixels,
            "state": self.state,
            "error": lambda _: self.error("not a valid query"),
        }

    def start(self):
        logger.info("Starting lit daemon")
        running = True
        serv = socket.socket(socket.AF_UNIX)
        try:
            os.remove(SOCKET_PATH)
        except OSError as e:
            logger.warning('Got "{}" when trying to remove {}'.format(e, SOCKET_PATH))
        try:
            # Allow created socket to have non-root read and write permissions
            os.umask(0o1)
            serv.bind(SOCKET_PATH)
            serv.listen(MAX_CONNECTIONS)
            logger.info("Listening on {}".format(SOCKET_PATH))
            while running:
                conn, address = serv.accept()
                self.start_conn_thread(conn)
        except KeyboardInterrupt:
            logger.warning("Shutting down due to keyboard interrupt")
            sys.exit(0)
        except Exception as e:
            logger.exception("litd socket error")
        finally:
            self.commands.stop_loop()

    def start_conn_thread(self, conn):
        def listener():
            while True:
                try:
                    data = conn.recv(4096)
                    if not data:
                        break
                    msg = data.decode()
                    logger.info("received command: {}".format(msg))
                    try:
                        resp = self.handle_request(msg).encode()
                    except Exception as e:
                        logger.exception("Unexpected error while handing command")
                        resp = self.error("Internal error").encode()
                    logger.debug("responding: {}".format(resp))
                    # First 32 bytes is message length
                    conn.send(str(len(resp)).zfill(32).encode())
                    conn.send(resp)
                except Exception as e:
                    logger.exception("litd connection error")

        thread = threading.Thread(target=listener)
        thread.start()

    def handle_request(self, data):
        msg = json.loads(data)
        type_error = self.error('type must be specified as "start", "stop", or "query"')
        if "start" in msg:
            return self.start_command(msg["start"])
        if "stop" in msg:
            return self.stop_command(msg["stop"])
        elif "query" in msg:
            return self.query(msg["query"])
        elif "dev" in msg:
            return self.dev_command(msg["dev"])
        else:
            return type_error

    @staticmethod
    def error(msg):
        return json.dumps({"code": 1, "message": "ERROR: {}".format(msg)})

    @staticmethod
    def result(data):
        data["code"] = 0
        return json.dumps(data)

    def stop_command(self, msg):
        if "effect_id" in msg:
            effect_id = msg["effect_id"]
            self.commands.stop_effect(effect_id)
        elif "transaction_id" in msg:
            transaction_id = msg["transaction_id"]
            for effect_id in self.commands.get_transaction(transaction_id):
                self.commands.stop_effect(effect_id)
        return json.dumps({"message": "Stopped", "code": 0})

    def start_command(self, msg):
        if "effect" in msg:
            transaction_id = next(TRANSACTION_IDS)
            effect = msg["effect"]
            response = self.commands.start_effect(
                effect["name"],
                effect.get("args", {}),
                effect.get("properties", {}),
                transaction_id,
            )
            return json.dumps(response.as_dict())
        elif "preset" in msg:
            transaction_id = next(TRANSACTION_IDS)
            preset = msg["preset"]
            response = self.commands.start_preset(
                preset["name"], preset.get("properties", {}), transaction_id
            )
            return json.dumps(response.as_dict())
        elif "history" in msg:
            direction = msg["history"]
            if direction.get("back"):
                success = self.commands.history.back()
            elif direction.get("forward"):
                success = self.commands.history.forward()
            else:
                return json.dumps(
                    {
                        "message": "History request must specify 'forward' or 'back'",
                        "code": 1,
                    }
                )
            if not success:
                return json.dumps({"message": "Nothing to do", "code": 5})
            return json.dumps(
                {
                    "message": "Went {}".format(
                        "back" if direction.get("back") else "forward"
                    ),
                    "code": 0,
                }
            )
        else:
            return json.dumps(
                {
                    "message": "Command must be one of 'effect', 'preset', or 'modify'",
                    "code": 1,
                }
            )

    def dev_command(self, msg):
        if msg["command"] == "verbosity":
            level = msg.get("args", {}).get("level", None)
            level_val = 0
            try:
                level_val = getattr(logging, level.upper())
            except AttributeError:
                return json.dumps(
                    {
                        "message": "the verbosity level must be debug, info, warning, error, or critical",
                        "code": 2,
                    }
                )
            logging.getLogger().setLevel(level_val)
            return json.dumps(
                {"message": "Logging level changed to {}".format(level), "code": 0}
            )
        return json.dumps({"message": "Unknown command", "code": 2})

    def query(self, msg):
        return self.queries[msg.get("what", "error")]()

    def effects(self):
        return self.result(
            {
                "effects": sorted(
                    self.commands.get_effects(), key=operator.itemgetter("name")
                )
            }
        )

    def presets(self):
        return self.result({"presets": sorted(self.commands.get_presets())})

    def colors(self):
        return self.result({"colors": self.commands.get_colors()})

    def color_types(self):
        return self.result({"color_types": self.commands.get_color_types()})

    def sections(self):
        return self.result({"sections": list(self.commands.get_sections())})

    def zones(self):
        return self.result({"zones": list(self.commands.get_zones())})

    def pixels(self):
        return self.result({"pixels": self.commands.get_pixels()})

    def state(self):
        return self.result({"state": self.commands.get_state()})


def start():
    try:
        litd = LitDaemon()
        litd.start()
    except Exception:
        logger.exception("Uncaught exception when starting daemon")


if __name__ == "__main__":
    logger.info("This module must be started by calling importing and calling start()")



import random

name = "Nexus"

start_message = name + " started!"

description = "Like the Nexus android wallpaper"


def gen_projectile(lights_size):
    right = random.random() < 0.5
    distance_from_start = random.randint(0, lights_size // 2)
    return [
        -distance_from_start
        if right
        else (lights_size - 1 + distance_from_start),  # Head position
        (1 if right else -1) * (max(0.3, random.normalvariate(2, 1))),  # Speed
        random.random(),  # Hue
    ]


def gen_starting_projectiles(lights, args):
    projectiles = []
    for _ in range(args["number"]):
        projectiles.append(gen_projectile(lights.size))
    return projectiles


schema = {
    "number": {
        "value": {
            "type": "number",  # TODO integer type (or step)
            "min": 1,
            "max": 50,
            "default": 10,
        },
        "user_input": True,
        "required": False,
        "index": 0,
    },
    "projectiles": {
        "value": {"type": "int list", "default_gen": gen_starting_projectiles},
        "user_input": False,
    },
}


def update(lights, step, state):
    projectiles = state["projectiles"]
    hsvs = [[0, 0, 0, 0] for _ in range(0, lights.size)]
    # Total of value(HSV) at each position
    count = [0] * lights.size
    for i, val in enumerate(projectiles):
        tail_length = int(15 * abs(val[1]))
        for t in range(0, tail_length):
            tail_pixel = int(val[0] + t * (1 if val[1] < 0 else -1))
            if tail_pixel >= 0 and tail_pixel < lights.size:
                (h, s, v) = (val[2], 1, (tail_length - 1.0 * t) / tail_length)
                hsvs[tail_pixel][0] = (
                    hsvs[tail_pixel][0] * count[tail_pixel] + (h * v)
                ) / (count[tail_pixel] + v)
                hsvs[tail_pixel][1] = s
                hsvs[tail_pixel][2] = max(hsvs[tail_pixel][2], v)
                hsvs[tail_pixel][3] = max(
                    hsvs[tail_pixel][3], (tail_length - 1.0 * t) / tail_length
                )
                count[tail_pixel] += v
        projectiles[i][0] = projectiles[i][0] + projectiles[i][1]
        if (val[0] + tail_length < 0 and val[1] < 0) or (
            val[0] - tail_length >= lights.size and val[1] > 0
        ):
            projectiles[i] = gen_projectile(lights.size)
    lights.set_pixels_hsv(hsvs)



import math

name = "Christmas"

start_message = name + " started!"

description = "Red and green pattern slides along the strand"


def setup_pixels(lights, args):
    colors = [None] * lights.size
    for n in range(0, lights.size):
        x = math.fabs((lights.size / 2 - n) / (lights.size / 2))
        colors[n] = (int(255 - (x * 255)), int(x * 255), 0)
    return colors


schema = {
    "speed": {
        "value": {"type": "number", "min": 0, "max": 100, "default": 50},
        "user_input": True,
        "required": False,
    },
    "colors": {
        "value": {"type": "array", "default_gen": setup_pixels},
        "user_input": False,
    },
}


def update(lights, step, state):
    state["colors"] = state["colors"][1:] + [state["colors"][0]]
    lights.set_pixels(state["colors"])



name = "Strobe"

start_message = name + " started!"

description = "Lights flash rapidly"

default_speed = 10

schema = {
    "color": {
        "value": {"type": "color", "default": (255, 255, 255)},
        "user_input": True,
        "required": False,
    },
}


def update(lights, step, state):
    color = state["color"].get_color(step)
    if step % 2 == 0:
        lights.set_all_pixels(color[0], color[1], color[2])
    else:
        lights.clear()



import random
import logging

logger = logging.getLogger(__name__)

name = "Twinkle"

start_message = name + " started!"

description = "Like a little star"

def setup_start_durations(lights, args):
    return [(-1, 1)] * lights.size


schema = {
    "color": {
        "value": {"type": "color", "default": (255, 255, 255)},
        "user_input": True,
        "required": False,
    },
    "twinkleyness": {
        "value": {"type": "number", "min": 1, "max": 200, "default": 50},
        "user_input": True,
        "required": False,
    },
    "brightnesses": {
        "value": {"type": "int list", "default_gen": lambda l, a: [0.5] * l.size},
        "user_input": False,
    },
}


def update(lights, step, state):
    brightnesses = state["brightnesses"]
    for i in range(lights.size):
        color = state["color"].get_color(step, i)
        brightness = brightnesses[i]
        lights.set_pixel(
            i,
            int(color[0] * brightness),
            int(color[1] * brightness),
            int(color[2] * brightness),
        )
        rand = random.random()
        if rand > brightness:
            brightnesses[i] += state["twinkleyness"] / 200
        elif rand > 1 - brightness:
            brightnesses[i] -= state["twinkleyness"] / 200
        brightnesses[i] = max(min(brightnesses[i], 1), 0)



class ColorType:
    def __init__(self, module):
        self.name = getattr(module, "name", "Unnamed")
        self.schema = getattr(module, "schema", {})
        self.module = module

    def new_generator(self, args, lights):
        return ColorGenerator(self, args, lights)

    def __str__(self):
        return str({"type": self.name, "schema": self.schema})


class ColorGenerator:
    def __init__(self, color_type, args, lights):
        self.generator_func = color_type.module.get_generator(args)
        self.args = args
        self.color_type = color_type
        self.lights_size = lights.size

    def get_color(self, step=0, position=0):
        return self.generator_func(step, position / self.lights_size)

    def get_palette(self, size):
        if hasattr(self.color_type.module, "get_palette"):
            return self.color_type.module.get_palette(size)
        return [self.get_color(i) for i in range(size)]

    def as_dict(self):
        return {"type": self.color_type.name, "args": self.args}

    def __str__(self):
        return str(self.as_dict)



name = "Police"

start_message = "Cops are here!"

description = "Mimics lights on top of police cars."

default_speed = 25

schema = {}

# CONSTANTS
LEFT = 0
CENTER_L = 2
CENTER_R = 3
RIGHT = 1


def update(lights, step, state):
    lights.clear()
    if step % 3 != 0:
        if (step // 10) % 2 == 0:
            set_section(lights, LEFT, (255, 0, 0))
            set_section(lights, CENTER_R, (255, 255, 255))
        else:
            set_section(lights, RIGHT, (0, 0, 255))
            set_section(lights, CENTER_L, (255, 255, 255))


def set_section(lights, section, color):
    if section == LEFT:
        for n in range(0, int(lights.size * (5 / 12))):
            lights.set_pixel(n, *color)
    elif section == RIGHT:
        for n in range(int(lights.size * (7 / 12)), lights.size):
            lights.set_pixel(n, *color)
    elif section == CENTER_L:
        for n in range(int(lights.size * (5 / 12)), int(lights.size * (1 / 2))):
            lights.set_pixel(n, *color)
    elif section == CENTER_R:
        for n in range(int(lights.size * (1 / 2)), int(lights.size * (7 / 12))):
            lights.set_pixel(n, *color)



import colorsys
import logging

logger = logging.getLogger(__name__)


class ControllerManager:
    def __init__(self, sections):
        """Creates a new Led_Controller Manager.
        sections: The named led sections that can be controlled. 
        """
        self.controllers = []
        self.sections = sections
        self.display_adapters = {
            s.section_adapter.display_adapter for s in self.sections.values()
        }
        self.section_ordered = sorted(
            self.sections, key=lambda e: self.sections[e].end_index
        )
        # Number of leds in all sections
        self.total_leds = sum(s.size for s in self.sections.values())
        # The rendered pixel values to be displayed
        self.pixels = [(0, 0, 0) for _ in range(self.total_leds)]
        # A mapping from absolute index to (local index, controller)
        self.pixel_locations = [None] * self.total_leds
        for section_name in self.section_ordered:
            section = self.sections[section_name]
            for local_i, absolute_i in enumerate(section.absolute_range):
                # TODO named tuple for pixel locations
                self.pixel_locations[absolute_i] = (local_i, section)

    def create_controller(self, sections, overlayed=False, opacity=1):
        """ Creates a new Led_Controller for the specified sections
        If overlayed is False , the sections will be removed from any existing controllers """
        controller = Controller(self, opacity)
        controller.set_sections(sections)
        if not overlayed:
            for c in self.controllers:
                if any(s in sections for s in c.active_section_names):
                    c.set_sections(
                        [s for s in c.active_section_names if s not in sections]
                    )
            # Remove any empty controllers
            self.controllers = [c for c in self.controllers if c.size != 0]
        self.controllers.append(controller)
        return controller

    def remove_controller(self, controller):
        self.controllers.remove(controller)

    def get_controllers(self):
        return self.controllers

    def render(self):
        overlay_pixels = [[] for _ in range(self.total_leds)]
        for controller in self.controllers:
            for abs_i, pixel in controller.abs_pixels():
                r, g, b, a = pixel

                overlay_pixels[abs_i].append((r, g, b, a * controller.opacity))
        for i, overlays in enumerate(overlay_pixels):
            r, g, b, a = 0, 0, 0, 0
            for pixel in overlays:
                if pixel[3] < 0:
                    a = -1
                    r = pixel[0]
                    g = pixel[1]
                    b = pixel[2]
                    break
                else:
                    a += pixel[3]
                    r += pixel[0] * pixel[3]
                    g += pixel[1] * pixel[3]
                    b += pixel[2] * pixel[3]
            if a < 0:
                self.pixels[i] = (r, g, b)
            elif a == 0:
                self.pixels[i] = (0, 0, 0)
            else:
                self.pixels[i] = (int(r // a), int(g // a), int(b // a))
            location = self.pixel_locations[i]
            location[1].section_adapter.set_pixel_color_rgb(
                location[0], *self.pixels[i]
            )

    def show(self):
        """Pushes the led array to the actual lights"""
        self.render()
        for display_adapter in self.display_adapters:
            display_adapter.show()

    def get_pixels(self):
        """ Returns the list of rgb values as of the last 'show' call """
        return self.pixels

    def __str__(self):
        return "Controller Manager: {}".format(self.controllers)


class Controller:
    """ A grouping of sections that are being controlled as one unit """

    def __init__(self, manager, opacity):
        """Creates a new Led_Controller with no active ranges"""
        self.sections = manager.sections
        self.section_ordered = sorted(
            self.sections, key=lambda e: self.sections[e].end_index
        )
        # Currently active section names sorted by end location
        self.active_section_names = []
        # Number of currently active leds
        self.size = 0
        self.opacity = opacity

    def set_sections(self, new_sections):
        """Sets the currently active sections to new_sections and updates other dependent fields.
        new_sections is a list of section names corresponding to the sections dict keys"""
        self.active_section_names = sorted(
            new_sections, key=lambda e: self.sections[e].end_index
        )
        self.size = sum(self.sections[r].size for r in self.active_section_names)
        self.pixels = [(0, 0, 0, 0)] * self.size

    def abs_pixels(self):
        controller_i = 0
        for sn in self.active_section_names:
            s = self.sections[sn]
            for abs_i in s.absolute_range:
                yield abs_i, self.pixels[controller_i]
                controller_i += 1

    def get_sections(self):
        """Returns a list of currently active sections"""
        return [self.sections[k] for k in self.active_section_names]

    def clear(self):
        """Clear the currently active buffer"""
        self.set_all_pixels(0, 0, 0, 0)

    def off(self):
        """ Turns off all pixels. """
        self.clear()

    def set_pixel(self, n, r, g, b, a=1):
        """Set a single pixel to RGB colour, with index only counting active sections for this controller"""
        self.pixels[n] = (r, g, b, a)

    def set_pixel_hsv(self, n, h, s, v, a=1):
        """Set a single pixel to RGB colour, with index only counting active sections.
        O(s) where s is the size of active sections"""
        r, g, b = [int(n * 255) for n in colorsys.hsv_to_rgb(h, s, v)]
        self.set_pixel(n, r, g, b, a)

    def set_pixels(self, pixels):
        """Set active pixels to corresponding pixels in array of rgb tuples with size 'size'"""
        for n in range(0, len(pixels)):
            if len(pixels[n]) == 4:
                r, g, b, a = pixels[n]
                self.set_pixel(n, r, g, b, a)
            else:
                r, g, b = pixels[n]
                self.set_pixel(n, r, g, b)

    def get_pixels(self):
        return self.pixels

    def set_pixels_hsv(self, pixels):
        """Set active pixels to corresponding pixels in array of hsv tuples with size 'size'"""
        for n in range(0, len(pixels)):
            r, g, b = [int(p * 255) for p in colorsys.hsv_to_rgb(*pixels[n][:3])]
            if len(pixels[n]) == 4:
                self.set_pixel(n, r, g, b, pixels[n][3])
            else:
                self.set_pixel(n, r, g, b)

    def set_all_pixels(self, r, g, b, a=1):
        """Sets all of the pixels to a color in the RGB colorspace"""
        self.pixels = [(r, g, b, a)] * self.size

    def set_all_pixels_hsv(self, h, s, v, a=1):
        """Sets all of the pixels to a color in the HSV colorspace"""
        r, g, b = [int(p * 255) for p in colorsys.hsv_to_rgb(h, s, v)]
        self.set_all_pixels(r, g, b, a)

    def __repr__(self):
        return "Controller: sections={} opacity={}".format(
            self.active_section_names, self.opacity
        )



import setuptools

setuptools.setup(
    name="lit-core",
    version="0.0.4",
    author="Nick Pesce",
    author_email="nickpesce22@gmail.com",
    description="Lit Ilumination Technology",
    url="https://github.com/lit-illumination-technology/lit_core",
    packages=["lit", "lit.effects", "lit.colors"],
    scripts=["bin/litctl", "bin/litdev"],
    entry_points={"console_scripts": ["litd=lit.litd:start"],},
    package_data={"lit": ["config/*"]},
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Topic :: Home Automation",
    ],
    install_requires=["rpi_ws281x"],
    python_requires=">=3.5",
)



import atexit
import colorsys
import importlib
import json
import logging
import os
import random
import sys
import time
import threading

from .controller import ControllerManager
from .error import InvalidEffectException
from .section import Section, SectionAdapter
from .device import DeviceAdapter
from . import effects
from . import colors
from .history import History
from .effect import Effect
from .color import ColorType

__author__ = "Nick Pesce"
__email__ = "nickpesce22@gmail.com"

# If the system is running slow, this many effect updates can happen before each light update
CATCHUP_UPDATES = 3
TIME_WARN_COOLDOWN = 10  # seconds
TIME_WARN_THRESHOLD = 0.1  # seconds
logger = logging.getLogger(__name__)


class commands:
    def __init__(self, base_path):
        self.base_path = base_path
        self.config_path = os.path.join(base_path, "config")
        logger.info("Using config directory: %s", self.config_path)
        self.sleep_event = threading.Event()

        self.color_types = {}
        self.effects = {}
        self.effects_by_id = {}
        self.transactions = {}
        self.sections = {}
        self.zones = {}
        self.default_range = None
        self.singleton_args = {}
        self.history = History(self)

        # Load config files
        with open(os.path.join(self.config_path, "presets.json")) as data_file:
            self.presets = json.load(data_file)

        with open(os.path.join(self.config_path, "colors.json")) as data_file:
            self.named_rgb_colors = json.load(data_file)

        with open(os.path.join(self.config_path, "ranges.json")) as data_file:
            range_json = json.load(data_file)
            section_json = range_json["sections"]
            adapters_json = range_json["adapters"]
            zone_json = range_json["zones"]
            self.default_range = range_json["default"]

            devices = {}
            for adapter in adapters_json:
                name = adapter["name"]
                if name in devices:
                    raise SyntaxError(
                        "Adapter name {name} was defined more than once. "
                        "Adapter names must be unique.".format(name=name)
                    )
                devices[name] = {
                    "adapter": DeviceAdapter.from_config(adapter),
                    "used_indexes": 0,
                }

            section_start_index = 0
            for section in section_json:
                device = devices.get(section["adapter"])
                if not device:
                    raise SyntaxError(
                        "Error in ranges.json: Section '{section}' references "
                        "adapter '{adapter}', but that adapter is not defined".format(
                            section=section["name"], adapter=section["adapter"]
                        )
                    )
                next_used_indexes = device["used_indexes"] + section["size"]
                if next_used_indexes > device["adapter"].size:
                    raise SyntaxError(
                        "Adapter '{name}' has {size} pixels available (adapter size), "
                        "but at least {used} were used by sections".format(
                            name=device["adapter"].name,
                            size=device["adapter"].size,
                            used=next_used_indexes,
                        )
                    )
                section_adapter = SectionAdapter(
                    device["used_indexes"], device["adapter"]
                )
                self.sections[section["name"]] = Section(
                    section["name"],
                    section_start_index,
                    section["size"],
                    section_adapter,
                )
                section_start_index += section["size"]
                device["used_indexes"] = next_used_indexes

            for zone in zone_json:
                self.zones[zone["name"]] = zone["sections"]

        self.controller_manager = ControllerManager(self.sections)
        self.import_colors()
        self.import_effects()
        atexit.register(self._clean_shutdown)
        self.start_loop()
        logger.info("Started effect loop")

    def start_loop(self):
        self.should_run = True
        self.show_lock = threading.Lock()

        def loop(self):
            total_steps = 0
            last_warn_time = 0
            while self.should_run:
                next_upd_time = float("inf")
                start_time = time.time()
                try:
                    with self.show_lock:
                        for effect in self.effects_by_id.values():
                            updates = 0
                            while (
                                effect.next_upd_time <= start_time
                                and updates < CATCHUP_UPDATES
                            ):
                                start_upd = time.time()
                                try:
                                    effect.update()
                                except Exception:
                                    logger.exception(
                                        'Error in effect "%s"', effect.effect.name
                                    )
                                end_upd = time.time()
                                logger.debug(
                                    "Took %dms to update %s",
                                    (end_upd - start_upd) * 1000,
                                    str(effect),
                                )
                                updates += 1
                            next_upd_time = min(next_upd_time, effect.next_upd_time)
                        logger.debug(
                            "Took %dms to do full update",
                            (time.time() - start_time) * 1000,
                        )
                        self.controller_manager.show()
                    if next_upd_time == float("inf"):
                        logger.debug("No effects updated")
                        continue

                    wait_time = next_upd_time - time.time()
                    if wait_time < -1 * TIME_WARN_THRESHOLD:
                        if last_warn_time < time.time() - TIME_WARN_COOLDOWN:
                            logger.warning(
                                "Can't keep up! Did the system time change, "
                                "or is the server overloaded? Running %d ms behind.",
                                (wait_time * -1000),
                            )
                            last_warn_time = time.time()
                    else:
                        self.sleep_event.wait(wait_time)
                    total_steps += 1
                except Exception as e:
                    if self.show_lock.locked():
                        self.show_lock.release()
                    logger.exception("Error in effect loop")

        self.loop_thread = threading.Thread(target=loop, args=(self,))
        self.loop_thread.start()

    def stop_loop(self):
        self.sleep_event.set()
        self.should_run = False
        self.loop_thread.join()

    def start_preset(self, preset_name, properties, transaction_id):
        preset = self.presets.get(preset_name, None)
        if not preset:
            msg = "The preset {} does not exist".format(preset_name)
            return Response(msg, 2)
        if "commands" not in preset:
            msg = "The preset {} does not specify commands".format(preset_name)
            return Response(msg, 3)
        for command in preset["commands"]:
            if "effect" not in command:
                msg = "The preset {} must specify an effect for all commands".format(
                    preset_name
                )
                return Response(msg, 3)
            effect_instance = self._start_effect(
                command["effect"],
                command.get("args", {}),
                command.get("properties", {}),
                transaction_id,
            )
            if not effect_instance:
                return Response("Invalid effect: {}".format(command["effect"]), 1)
        self.history.save(self.get_state())
        return Response(
            preset.get("start_message", "{} started!".format(preset_name)),
            0,
            transaction_id,
        )

    def start_effect(self, effect_name, args, properties, transaction_id):
        if not self.is_effect(effect_name):
            raise InvalidEffectException()
        effect_instance = self._start_effect(
            effect_name, args, properties, transaction_id
        )
        self.history.save(self.get_state())
        return Response(effect_instance.effect.start_message, 0, transaction_id)

    def _start_effect(self, effect_name, args, properties, transaction_id):
        if not self.is_effect(effect_name):
            return None
        effect = self.effects[effect_name.lower()]
        sections = self.get_sections_from_ranges(
            properties.get("ranges", [self.default_range])
        )
        with self.show_lock:
            opacity = properties.get("opacity")
            if opacity is None:
                opacity = 1
            if not isinstance(opacity, (int, float)):
                logger.warning("Invalid 'opacity' property received: %s", str(opacity))
                opacity = 1
            overlayed = properties.get("overlayed")
            if overlayed is None:
                overlayed = False
            if not isinstance(overlayed, bool):
                logger.warning(
                    "Invalid 'overlayed' property received: %s", str(overlayed)
                )
                overlayed = False
            controller = self.controller_manager.create_controller(
                sections, overlayed=overlayed, opacity=opacity
            )
            self.effects_by_id = {
                effect_id: effect
                for effect_id, effect in self.effects_by_id.items()
                if effect.controller.size > 0
            }

            schema = effect.schema
            # remove any 'None' args
            args = {k: v for (k, v) in args.items() if v is not None}
            # fill in default args from schema
            self.complete_args_with_schema(args, schema, controller)
            # attempt to parse arg values
            args = {k: self.process_arg_value(k, args[k], controller) for k in args}

            default_speed = effect.default_speed
            speed = properties.get("speed", default_speed)
            effect_instance = effect.create(args, speed, controller, transaction_id)
            self.transactions.setdefault(transaction_id, []).append(effect_instance.id)
            self.effects_by_id[effect_instance.id] = effect_instance
        self.sleep_event.set()
        logger.info("New controller manager: %s", str(self.controller_manager))
        return effect_instance

    def get_transaction(self, transaction_id):
        return self.transactions.get(transaction_id)

    def stop_all(self):
        with self.show_lock:
            effect_ids = [effect_id for effect_id in self.effects_by_id]
        for effect_id in effect_ids:
            self.stop_effect(effect_id)

    def stop_effect(self, effect_id):
        with self.show_lock:
            effect = self.effects_by_id.get(effect_id)
            if effect:
                self.controller_manager.remove_controller(effect.controller)
                del self.effects_by_id[effect_id]
                return True
            return False

    def complete_args_with_schema(self, args, schema, controller):
        for k, v in sorted(
            schema.items(), key=lambda x: x[1].get("index", float("inf"))
        ):
            if (not k in args or not v.get("user_input", False)) and (
                "default" in v["value"] or "default_gen" in v["value"]
            ):
                # If this arg is a singleton that was already set, give a reference
                if v.get("singleton", False) and k in self.singleton_args:
                    args[k] = self.singleton_args[k]
                # Otherwise, get the default value
                else:
                    if "default_gen" in v["value"]:
                        args[k] = v["value"]["default_gen"](controller, args)
                    else:
                        args[k] = v["value"]["default"]
                    if v.get("singleton", False):
                        self.singleton_args[k] = args[k]

    def help(self):
        return """Effects:\n    ~ """ + (
            "\n    ~ ".join(
                effect.name + " " + self.schema_to_string(effect.command_schema)
                for effect in self.effects.values()
            )
        )

    def schema_to_string(self, schema):
        ret = str(schema)
        return ret

    def get_effects(self):
        return [
            {
                "name": effect.name,
                "schema": effect.command_schema,
                "default_speed": effect.default_speed,
            }
            for effect in self.effects.values()
        ]

    def get_presets(self):
        return self.presets

    def get_colors(self):
        return self.named_rgb_colors

    def get_color_types(self):
        return [
            {"name": ct.name, "schema": ct.schema} for ct in self.color_types.values()
        ]

    def get_sections(self):
        return self.sections

    def get_zones(self):
        return self.zones

    def get_pixels(self):
        return self.controller_manager.get_pixels()

    def get_state(self):
        state = []

        def state_schema(state, schema):
            processed = {}
            for k, v in state.items():
                if k in schema:
                    if (
                        not isinstance(v, (dict, str, int, float, bool))
                        and v is not None
                    ):
                        try:
                            v = v.as_dict()
                        except AttributeError:
                            v = str(v)
                    processed[k] = v
            return processed

        with self.show_lock:
            for effect in self.effects_by_id.values():
                controller = effect.controller
                state.append(
                    {
                        "sections": controller.active_section_names,
                        "opacity": controller.opacity,
                        "effect_name": effect.effect.name,
                        "effect_state": state_schema(
                            effect.state, effect.effect.command_schema
                        ),
                        "effect_id": effect.id,
                        "transaction_id": effect.transaction_id,
                    }
                )
        return state

    def get_sections_from_ranges(self, lst):
        """converts ranges names (sections or zones), to a list containing ony section names"""
        ret = []
        for r in lst:
            if r in self.sections:
                ret.append(r)
            elif r in self.zones:
                ret += self.zones[r]
        return ret

    def process_arg_value(self, arg_type, value, lights_controller):
        """Given a attribute represented as a string, convert it to the appropriate value"""
        if arg_type.lower() == "color":
            generator_schema = None
            rgb = None
            if isinstance(value, dict):
                # Color generator
                generator_schema = value
            elif isinstance(value, str):
                if value == "random":
                    rgb = list(
                        map(
                            lambda x: int(255 * x),
                            colorsys.hsv_to_rgb(random.random(), 1, 1),
                        )
                    )
                else:
                    for color in self.named_rgb_colors:
                        if color["name"].lower() == value.lower():
                            rgb = color["rgb"]
                    raise ValueError(f"Invalid color name: {value}")
            elif isinstance(value, (tuple, list)):
                # RGB
                rgb = list(value)
            else:
                raise ValueError(f"Invalid type for color argument: {type(value)}")
            if not generator_schema:
                generator_schema = {
                    "type": "Static Color",
                    "args": {"color": rgb},
                }
            color_type = self.color_types.get(generator_schema["type"].lower())
            if not color_type:
                raise ValueError(
                    f"Invalid color generator type: {generator_schema.get('type')}"
                )
            return color_type.new_generator(generator_schema["args"], lights_controller)
        return value

    def is_effect(self, name):
        return name.lower() in self.effects

    def import_effects(self):
        module_names = effects.effects
        modules = []
        for m in module_names:
            try:
                module = importlib.import_module(m)
                modules.append(module)
            except Exception as e:
                logger.exception("Could not import %s", m)

        logger.info("Parsing effect modules")
        for module in modules:
            try:
                logger.info("loading %s", str(module))
                effect = Effect(module)
                self.effects[effect.name.lower()] = effect
            except Exception as e:
                logger.exception("Error loading effect %s", str(module))

    def import_colors(self):
        module_names = colors.colors
        modules = []
        for m in module_names:
            try:
                module = importlib.import_module(m)
                modules.append(module)
            except Exception as e:
                logger.exception("Could not import %s", m)

        logger.info("Parsing color generator modules")
        for module in modules:
            try:
                logger.info("loading %s", str(module))
                color_type = ColorType(module)
                self.color_types[color_type.name.lower()] = color_type
            except Exception as e:
                logger.exception("Error loading color type %s", str(module))

    def _clean_shutdown(self):
        logger.info("Shutting down")
        self.stop_loop()
        for controller in self.controller_manager.get_controllers():
            controller.clear()
        self.controller_manager.show()


class Response:
    def __init__(self, message, rc, transaction_id=None):
        self.message = message
        self.rc = rc
        self.transaction_id = transaction_id

    def as_dict(self):
        return {
            "message": self.message,
            "code": self.rc,
            "transaction_id": self.transaction_id,
        }


if __name__ == "__main__":
    logger.error("This is module can not be run. Import it and call start()")
    sys.exit()



import colorsys
import random

name = "Cycle"
description = "Cycle through the spectrum of colors"
schema = {}


def get_generator(args):
    def generator(step, _position):
        return tuple(
            map(lambda x: int(255 * x), colorsys.hsv_to_rgb((step % 255) / 255, 1, 1))
        )

    return generator



import random

name = "Rave"

start_message = name + " started!"

description = "Each light flashes a random color"

schema = {}


def update(lights, step, state):
    for i in range(lights.size):
        lights.set_pixel(
            i, random.randint(0, 255), random.randint(0, 255), random.randint(0, 255)
        )



import os
import glob

effects_path = os.path.dirname(os.path.abspath(__file__))
modules = glob.glob(os.path.join(effects_path, "**", "*.py"), recursive=True)
relative_paths = [f[len(effects_path) + 1 : -3] for f in modules if os.path.isfile(f)]
split_paths = [os.path.normpath(p).split(os.sep) for p in relative_paths]
effects = ["lit.effects.{}".format(".".join(p)) for p in split_paths if p[-1][0] != "_"]



name = "Christmas Lights"

start_message = name + " started!"

description = "Colors of traditional colored christmas lights"

schema = {}

BRIGHTNESS = 0.2


def update(lights, step, state):
    for i in range(lights.size):
        seq = i % 10
        if seq == 0:
            lights.set_pixel(i, int(170 * BRIGHTNESS), 0, 0)
        elif seq == 2:
            lights.set_pixel(i, int(170 * BRIGHTNESS), 0, int(85 * BRIGHTNESS))
        elif seq == 4:
            lights.set_pixel(i, 0, int(170 * BRIGHTNESS), 0)
        elif seq == 6:
            lights.set_pixel(i, int(255 * BRIGHTNESS), int(170 * BRIGHTNESS), 0)
        elif seq == 8:
            lights.set_pixel(i, 0, 0, int(170 * BRIGHTNESS))
        else:
            lights.set_pixel(i, 0, 0, 0)



import random

name = "Drip"

start_message = name + " started!"

description = "Mimics water droplets accumulating and falling"


def setup_dullness(lights, args):
    return [0] * lights.size


schema = {
    "color": {
        "value": {"type": "color", "default": (0, 20, 175)},
        "user_input": True,
        "required": False,
    },
    "intensity": {
        "value": {"type": "number", "min": 1, "max": 100, "default": 50},
        "user_input": True,
        "required": False,
    },
    "dullness": {
        "value": {
            "type": "number list",
            "default_gen": (lambda lights, args: [0] * lights.size),
        },
        "user_input": False,
    },
    "colors": {
        "value": {
            "type": "color list",
            "default_gen": (lambda lights, args: [(0, 0, 0)] * lights.size),
        },
        "user_input": False,
    },
}


def update(lights, step, state):
    dullness = state["dullness"]
    colors = state["colors"]
    intensity = state["intensity"]
    for i in range(lights.size):
        if dullness[i] <= 1 or random.randint(0, int(dullness[i] ** 2)) == 0:
            dullness[i] = (1000 / intensity) * (random.random() + 1)
            colors[i] = state["color"].get_color(step, i)
        lights.set_pixel(
            i,
            int(colors[i][0] / dullness[i]),
            int(colors[i][1] / dullness[i]),
            int(colors[i][2] / dullness[i]),
        )
        dullness[i] -= (random.random() * intensity) / 100
